@@ -1,5 +1,5 @@
 """
-    Homonim: Radiometric homogenisation of aerial and satellite imagery
+    Homonim: Correction of aerial and satellite imagery to surface reflectance
     Copyright (C) 2021 Dugal Harris
     Email: dugalh@gmail.com
 
@@ -17,105 +17,130 @@
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 """
 
-from typing import Tuple
+from typing import Tuple, Dict, Optional
 
 import cv2 as cv
 import numpy as np
-from homonim import utils
-from homonim.enums import Method
-from homonim.raster_array import RasterArray
 from rasterio.enums import Resampling
 from rasterio.fill import fillnodata
 
+from homonim import utils
+from homonim.enums import Model
+from homonim.raster_array import RasterArray
+
+ONdArray = Optional[np.ndarray]
+OShape = Optional[Tuple[int, int]]
+
 
 class KernelModel:
-    """
-    A base class for surface reflectance modelling and homogenisation of blocks of image data.
 
-    The surface reflectance relationship between source and reference image blocks is approximated with localised linear
-    models.  Models are estimated for each pixel location inside a small rectangular kernel (window), using a fast DFT
-    approach.  The homogenised output is produced by applying the model parameters to the source image.
-
-    Based on the paper:
-    Harris, Dugal & Van Niekerk, Adriaan. (2018). Radiometric homogenisation of aerial images by calibrating with
-    satellite data. International Journal of Remote Sensing. 40. 1-25. 10.1080/01431161.2018.1528404.
-    https://www.researchgate.net/publication/328317307_Radiometric_homogenisation_of_aerial_images_by_calibrating_with_satellite_data
-    """
-
-    default_config = dict(downsampling='average', upsampling='cubic_spline', r2_inpaint_thresh=0.25,
-                          mask_partial=False)
-
-    def __init__(self, method, kernel_shape, param_image=False, r2_inpaint_thresh=default_config['r2_inpaint_thresh'],
-                 mask_partial=default_config['mask_partial'], downsampling=default_config['downsampling'],
-                 upsampling=default_config['upsampling']):
+    def __init__(self, model: Model, kernel_shape: Tuple[int, int], find_r2: bool = False, **kwargs):
         """
-        Construct a KernelModel.
+        A base class for surface reflectance modelling and correction of blocks of image data.
+
+        The surface reflectance relationship between source and reference image blocks is approximated with localised
+        linear models.  Models are estimated for each pixel location inside a small rectangular kernel (window),
+        using a fast DFT approach.  The corrected output is produced by applying the model parameters to the source
+        image block.
+
+        Based on the paper:
+        Harris, Dugal & Van Niekerk, Adriaan. (2018). Radiometric homogenisation of aerial images by calibrating with
+        satellite data. International Journal of Remote Sensing. 40. 1-25. 10.1080/01431161.2018.1528404.
+        https://www.researchgate.net/publication
+        /328317307_Radiometric_homogenisation_of_aerial_images_by_calibrating_with_satellite_data
 
         Parameters
         ----------
-        method: homonim.enums.Method
-                The radiometric homogenisation method.
+        model: homonim.enums.Model
+            Surface reflectance correction model.
         kernel_shape: tuple
-                The (height, width) of the kernel in pixels.
-        param_image: bool, optional
-                Turn on/off the calculation of Pearson's correlation coefficient (r2) for each kernel model.
-                (Typically used for producing 'debug' images of the model parameters and accuracies.)
-                [default: False]
-        r2_inpaint_thresh: float, optional
-                The R2 (coefficient of determination) threshold below which to 'in-paint' kernel model parameters from
-                surrounding areas (applies to 'method' == Method.gain_offset only).  For pixels where the model gives a
-                poor approximation to the data (this can occur in e.g. shadowed areas, poor source-reference
-                co-registration, or areas where there has been land cover change between the source and reference
-                images), model offsets are interpolated from surrounding areas, and gains re-estimated.  It can be set
-                to 'None' to turn off in-painting.
-                [default: 0.25]
-        mask_partial: bool, optional
-                Masks output pixels that were not produced by full kernel or source/reference image coverage.  Useful
-                for ensuring strict model validity, and reducing seam-lines between adjacent images. [default: False]
-        downsampling: rasterio.enums.Resampling
-                The resampling method to use when downsampling.  [default: Resampling.average]
-        upsampling: rasterio.enums.Resampling
-                The resampling method to use when upsampling.  [default: Resampling.cubic_spline]
+            (height, width) of the kernel in pixels.
+        find_r2: bool, optional
+            Whether to calculate *R*\\ :sup:`2` (coefficient of determination) for each kernel model, and include in
+            parameter arrays.
+        kwargs:
+            Optional configuration arguments.  See :meth:`KernelModel.create_config` for keys and defaults values.
         """
+        self._model = Model(model)
+        self._kernel_shape = utils.validate_kernel_shape(kernel_shape, model=model)
+        self._find_r2 = find_r2
 
-        if not isinstance(method, Method):
-            raise ValueError("'method' should be an instance of homonim.enums.Method")
-        self._method = method
-
-        self._kernel_shape = utils.validate_kernel_shape(kernel_shape, method=method)
-
-        self._param_image = param_image
-        self._r2_inpaint_thresh = r2_inpaint_thresh
-        self._mask_partial = mask_partial
-        self._downsampling = downsampling
-        self._upsampling = upsampling
+        # update config defaults with any passed values, and set attributes
+        config = self.create_config(**kwargs)
+        self._r2_inpaint_thresh: float = config['r2_inpaint_thresh']
+        self._mask_partial: bool = config['mask_partial']
+        self._downsampling: Resampling = config['downsampling']
+        self._upsampling: Resampling = config['upsampling']
 
     @property
-    def method(self) -> Method:
-        """The homogenisation method."""
-        return self._method
+    def model(self) -> Model:
+        """ Correction model type. """
+        return self._model
 
     @property
     def kernel_shape(self) -> Tuple[int, int]:
-        """The kernel (height, width)."""
+        """ Kernel (height, width) in pixels. """
         return tuple(self._kernel_shape)
 
     @property
-    def config(self) -> dict:
-        """A dict containing the KernelModel configuration."""
-        return dict(r2_inpaint_thresh=self._r2_inpaint_thresh, downsampling=self._downsampling,
-                    upsampling=self._upsampling, mask_partial=self._mask_partial)
+    def find_r2(self) -> bool:
+        """ Whether to create and include R2 (coefficient of determination) in parameter arrays. """
+        return self._find_r2
 
-    def _get_resampling(self, from_res, to_res):
-        """Get the resampling method for re-projecting from resolution `from_res` to resolution `to_res`"""
+    @staticmethod
+    def create_config(
+        r2_inpaint_thresh: float = 0.25, mask_partial: bool = False, downsampling: Resampling = Resampling.average,
+        upsampling: Resampling = Resampling.cubic_spline,
+    ) -> Dict:
+        """
+        Utility method to create a KernelModel configuration dictionary that can be passed to
+        :meth:`KernelModel.__init__` and :meth:`homonim.RasterFuse.process`.  Without arguments, the default
+        configuration values are returned.
+
+        Parameters
+        ----------
+        r2_inpaint_thresh: float, optional
+            *R*\\ :sup:`2` (coefficient of determination) threshold below which to `in-paint` kernel model parameters
+            from surrounding areas (applies to the :attr:`~homonim.enums.Model.gain_offset` model only).  For
+            pixels where the model gives a poor approximation to the data (this can occur in areas where source and
+            reference differ due to e.g. shadowing, land cover changes etc.), model offsets are interpolated from
+            surrounding areas, and gains re-estimated.  Set ``r2_inpaint_thresh`` to `None` to turn off in-painting.
+        mask_partial: bool, optional
+            Mask output pixels not produced by full kernel or source/reference image coverage.  Useful for ensuring
+            strict model validity, and reducing seam-lines between overlapping images.
+        downsampling: rasterio.enums.Resampling, optional
+            Resampling method to use when downsampling.  See the `rasterio docs
+            <https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for
+            available options.
+        upsampling: rasterio.enums.Resampling, optional
+            Resampling method to use when upsampling. See the `rasterio docs
+            <https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for
+            available options.
+
+        Returns
+        -------
+        dict
+            Configuration dictionary.
+        """
+        return dict(
+            r2_inpaint_thresh=r2_inpaint_thresh, mask_partial=mask_partial, downsampling=downsampling,
+            upsampling=upsampling,
+        )
+
+    def _get_resampling(self, from_res: Tuple[float, float], to_res: Tuple[float, float]):
+        """ Return the resampling method for re-projecting from resolution ``from_res`` to resolution ``to_res``. """
         return self._downsampling if np.prod(np.abs(from_res)) <= np.prod(np.abs(to_res)) else self._upsampling
 
-    def _r2_array(self, ref_array, src_array, param_array, mask=None, mask_sum=None, ref_sum=None, src_sum=None,
-                  ref2_sum=None, src2_sum=None, src_ref_sum=None, dest_array=None, kernel_shape=None):
+    def _r2_array(
+        self, ref_array: np.ndarray, src_array: np.ndarray, param_array: np.ndarray, mask: ONdArray = None,
+        mask_sum: ONdArray = None, ref_sum: ONdArray = None, src_sum: ONdArray = None, ref2_sum: ONdArray = None,
+        src2_sum: ONdArray = None, src_ref_sum: ONdArray = None, dest_array: ONdArray = None,
+        kernel_shape: OShape = None
+    ) -> np.ndarray:
         """
-        Helper function to find R2 coefficient of determination at each pixel/kernel location for the given matrices.
+        Utility function to return R2 (coefficient of determination) at each pixel/kernel location for the given
+        arrays.
         """
-
         if kernel_shape is None:
             kernel_shape = self._kernel_shape
         kernel_shape = tuple(kernel_shape)  # force to tuple for opencv
@@ -123,9 +148,11 @@ class KernelModel:
 
         # find the keyword arguments that were not provided
         if mask is None:
-            # if mask is passed, assume that it has been applied ref_array and src_array, otherwise do that here
-            mask = (~utils.nan_equals(src_array, RasterArray.default_nodata) &
-                    ~utils.nan_equals(ref_array, RasterArray.default_nodata))
+            # if mask is passed, assume that it has been applied to ref_array and src_array, otherwise do that here
+            mask = (
+                ~utils.nan_equals(src_array, RasterArray.default_nodata) &
+                ~utils.nan_equals(ref_array, RasterArray.default_nodata)
+            )
             ref_array[~mask] = 0
             src_array[~mask] = 0
         if mask_sum is None:
@@ -140,26 +167,27 @@ class KernelModel:
             src_ref_sum = cv.boxFilter(src_array * ref_array, -1, kernel_shape[::-1], **filter_args)
 
         # R2 is found using: R2 = 1 - (residual sum of squares)/(total sum of squares) = 1 - RSS/TSS
-
         # TSS = sum((ref - mean(ref))**2), which can be expanded and expressed in terms of cv.boxFilter kernel sums as:
         ss_tot_array = (mask_sum * ref2_sum) - (ref_sum ** 2)
 
         if param_array.shape[0] > 1:
-            # find RSS for method == Method.gain_offset
+            # find RSS for model == Model.gain_offset
             if src_sum is None:
                 src_sum = cv.boxFilter(src_array, -1, kernel_shape[::-1], **filter_args)
 
             # RSS = sum((ref - ref_hat)**2)
             #     = sum((ref - (m*src + c))**2), where m and c are the first 2 bands of param_array
             # The above can be expanded and expressed in terms of cv.boxFilter kernel sums as:
-
-            ss_res_array = (((param_array[0] ** 2) * src2_sum) +
-                            (2 * np.product(param_array[:2], axis=0) * src_sum) -
-                            (2 * param_array[0] * src_ref_sum) -
-                            (2 * param_array[1] * ref_sum) +
-                            ref2_sum + (mask_sum * (param_array[1] ** 2)))
+            ss_res_array = (
+                ((param_array[0] ** 2) * src2_sum) +
+                (2 * np.product(param_array[:2], axis=0) * src_sum) -
+                (2 * param_array[0] * src_ref_sum) -
+                (2 * param_array[1] * ref_sum) +
+                ref2_sum + (mask_sum * (param_array[1] ** 2))
+            )  # yapf: disable
         else:
-            # find RSS for method == Method.gain or Method.gain_blk_offset
+            # find RSS for model == Model.gain or Model.gain_blk_offset
+
             # RSS = sum((ref - m*src)**2), where m is the first band of param_array
             # The above can be expanded and expressed in terms of cv.boxFilter kernel sums as:
             ss_res_array = (((param_array[0] ** 2) * src2_sum) - (2 * param_array[0] * src_ref_sum) + ref2_sum)
@@ -168,8 +196,9 @@ class KernelModel:
 
         if dest_array is None:
             # assign a destination array to write R2 into, if it was not provided
-            dest_array = np.full(src_array.shape, fill_value=RasterArray.default_nodata,
-                                 dtype=RasterArray.default_dtype)
+            dest_array = np.full(
+                src_array.shape, fill_value=RasterArray.default_nodata, dtype=RasterArray.default_dtype
+            )
 
         # find R2 = 1 - RSS/TSS, and write into dest_array
         np.divide(ss_res_array, ss_tot_array, out=dest_array, where=mask)
@@ -177,22 +206,11 @@ class KernelModel:
         return dest_array
 
     @staticmethod
-    def _fit_block_norm(ref_ra, src_ra):
+    def _fit_block_norm(src_ra: RasterArray, ref_ra: RasterArray) -> np.ndarray:
         """
-        Find a scale and offset for a source block, so that the standard deviation and first percentile of the source
-        and reference blocks match.  (Can be thought of as a rough dark object subtraction (DOS) approach).
-
-        Parameters
-        ----------
-        ref_ra : RasterArray
-            Reference data block in a RasterArray.
-        src_ra : RasterArray
-            Source data block in a RasterArray, with the same CRS, shape & extents as ref_ra.
-
-        Returns
-        -------
-        norm_model: numpy.array
-            A two element [gain, offset] model to 'normalise' the source block.
+        Return a two element [gain, offset] model to 'normalise' the source block, so that the standard deviation and
+        first percentile of the source and reference blocks match.  (Can be thought of as a basic dark object
+        subtraction (DOS) approach).
         """
         norm_model = np.zeros(2)
         mask = ref_ra.mask & src_ra.mask
@@ -202,24 +220,11 @@ class KernelModel:
         norm_model[1] = np.percentile(ref_ra.array[mask], 1) - np.percentile(src_ra.array[mask], 1) * norm_model[0]
         return norm_model
 
-    def _fit_gain(self, ref_ra, src_ra, kernel_shape=None):
+    def _fit_gain(self, src_ra: RasterArray, ref_ra: RasterArray, kernel_shape: OShape = None) -> RasterArray:
         """
-        Find sliding kernel gains, for a band, using opencv convolution.
-
-        Parameters
-        ----------
-        ref_ra : RasterArray
-            Reference data block in a RasterArray.
-        src_ra : RasterArray
-            Source data block in a RasterArray, with the same CRS, shape & extents as ref_ra.
-        kernel_shape : tuple, optional
-            Sliding kernel [height, width] in pixels.
-
-        Returns
-        -------
-        param_ra :RasterArray
-            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
-            R2 for each kernel model in the third band when param_image==True.
+        Find sliding kernel gains, for a source & reference band, using opencv convolution.
+        Returns a RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second,
+        and optionally R2 for each kernel model in the third band when :attr:`find_r2` is True.
         """
         # adapted from https://www.mathsisfun.com/data/least-squares-regression.html with c=0
         if kernel_shape is None:
@@ -233,10 +238,11 @@ class KernelModel:
         ref_array[~mask] = 0
         src_array[~mask] = 0
 
-        # setup a RasterArray profile for the parameters
+        # set up a RasterArray profile for the parameters
         param_profile = src_ra.profile.copy()
-        param_profile.update(count=3 if self._param_image else 2, nodata=RasterArray.default_nodata,
-                             dtype=RasterArray.default_dtype)
+        param_profile.update(
+            count=3 if self._find_r2 else 2, nodata=RasterArray.default_nodata, dtype=RasterArray.default_dtype
+        )
 
         # convolve the kernel with src and ref to get kernel sums (uses DFT for large kernels)
         filter_args = dict(normalize=False, borderType=cv.BORDER_CONSTANT)  # common opencv arguments
@@ -250,36 +256,29 @@ class KernelModel:
         # find sliding kernel gains, avoiding divide by 0
         np.divide(ref_sum, src_sum, out=param_ra.array[0], where=mask)
 
-        if self._param_image:
+        if self._find_r2:
             # Find R2 of the sliding kernel models
-            self._r2_array(ref_array, src_array, param_ra.array[:1], mask=mask, ref_sum=ref_sum, src_sum=src_sum,
-                           dest_array=param_ra.array[2], kernel_shape=kernel_shape)
+            self._r2_array(
+                ref_array, src_array, param_ra.array[:1], mask=mask, ref_sum=ref_sum, src_sum=src_sum,
+                dest_array=param_ra.array[2], kernel_shape=kernel_shape
+            )
 
         return param_ra
 
-    def _fit_gain_blk_offset(self, ref_ra, src_ra, kernel_shape=None):
+    def _fit_gain_blk_offset(
+        self, src_ra: RasterArray, ref_ra: RasterArray, kernel_shape: OShape = None
+    ) -> RasterArray:  # yapf: disable
         """
         Find sliding kernel gains and 'image' (i.e. block) offset, for a band, using opencv convolution.
-
-        ref_ra : RasterArray
-            Reference data block in a RasterArray.
-        src_ra : RasterArray
-            Source data block in a RasterArray, with the same CRS, shape & extents as ref_ra.
-        kernel_shape : tuple, optional
-            Sliding kernel [height, width] in pixels.
-
-        Returns
-        -------
-        param_ra :RasterArray
-            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
-            R2 for each kernel model in the third band when param_image==True.
+        Returns a RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second,
+        and optionally R2 for each kernel model in the third band when :attr:`find_r2` is True.
         """
         if kernel_shape is None:
             kernel_shape = self._kernel_shape
         kernel_shape = tuple(kernel_shape)  # force to tuple for opencv
 
         # find the source->reference normalisation
-        norm_model = self._fit_block_norm(ref_ra, src_ra)
+        norm_model = self._fit_block_norm(src_ra, ref_ra)
 
         # force src nodata to nan so that operation below remains correctly masked
         src_ra.nodata = RasterArray.default_nodata
@@ -288,29 +287,18 @@ class KernelModel:
         src_ra.array = (src_ra.array * norm_model[0]) + norm_model[1]
 
         # find gains for normalised source
-        param_ra = self._fit_gain(ref_ra, src_ra, kernel_shape=kernel_shape)
+        param_ra = self._fit_gain(src_ra, ref_ra, kernel_shape=kernel_shape)
 
         # incorporate the normalisation model in the parameter RasterArray
         param_ra.array[1] = param_ra.array[0] * norm_model[1]
         param_ra.array[0] *= norm_model[0]
         return param_ra
 
-    def _fit_gain_offset(self, ref_ra, src_ra, kernel_shape=None):
+    def _fit_gain_offset(self, src_ra: RasterArray, ref_ra: RasterArray, kernel_shape: OShape = None) -> RasterArray:
         """
         Find sliding kernel full linear model for a band using opencv convolution.
-
-        ref_ra : RasterArray
-            Reference data block in a RasterArray.
-        src_ra : RasterArray
-            Source data block in a RasterArray, with the same CRS, shape & extents as ref_ra.
-        kernel_shape : tuple, optional
-            Sliding kernel [height, width] in pixels.
-
-        Returns
-        -------
-        param_ra :RasterArray
-            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
-            R2 for each kernel model in the third band when param_image==True.
+        Returns a RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second,
+        and optionally R2 for each kernel model in the third band when :attr:`find_r2` is True.
         """
         # Least squares formulae adapted from https://www.mathsisfun.com/data/least-squares-regression.html
         if kernel_shape is None:
@@ -324,18 +312,21 @@ class KernelModel:
         ref_array[~mask] = 0
         src_array[~mask] = 0
 
-        # setup a RasterArray profile for the parameters
+        # set up a RasterArray profile for the parameters
         param_profile = src_ra.profile.copy()
-        find_r2 = self._param_image or (self._r2_inpaint_thresh is not None)
-        param_profile.update(count=3 if find_r2 else 2, nodata=RasterArray.default_nodata,
-                             dtype=RasterArray.default_dtype)
+        find_r2 = self._find_r2 or (self._r2_inpaint_thresh is not None)
+        param_profile.update(
+            count=3 if find_r2 else 2, nodata=RasterArray.default_nodata, dtype=RasterArray.default_dtype
+        )
 
         # find the numerator for the gain i.e N*cov(ref, src)
         filter_args = dict(normalize=False, borderType=cv.BORDER_CONSTANT)  # common opencv arguments
         src_sum = cv.boxFilter(src_array, -1, kernel_shape[::-1], **filter_args)
         ref_sum = cv.boxFilter(ref_array, -1, kernel_shape[::-1], **filter_args)
         src_ref_sum = cv.boxFilter(src_array * ref_array, -1, kernel_shape[::-1], **filter_args)
-        mask_sum = cv.boxFilter(mask.astype(RasterArray.default_dtype, copy=False), -1, kernel_shape[::-1], **filter_args)
+        mask_sum = cv.boxFilter(
+            mask.astype(RasterArray.default_dtype, copy=False), -1, kernel_shape[::-1], **filter_args
+        )
         m_num_array = (mask_sum * src_ref_sum) - (src_sum * ref_sum)
 
         # find the denominator for the gain i.e. N*var(src)
@@ -353,9 +344,11 @@ class KernelModel:
 
         if find_r2:
             # Find R2 of the sliding kernel models
-            self._r2_array(ref_array, src_array, param_ra.array[:2], mask=mask, mask_sum=mask_sum, ref_sum=ref_sum,
-                           src_sum=src_sum, src2_sum=src2_sum, src_ref_sum=src_ref_sum, dest_array=param_ra.array[2],
-                           kernel_shape=kernel_shape)
+            self._r2_array(
+                ref_array, src_array, param_ra.array[:2], mask=mask, mask_sum=mask_sum, ref_sum=ref_sum,
+                src_sum=src_sum, src2_sum=src2_sum, src_ref_sum=src_ref_sum, dest_array=param_ra.array[2],
+                kernel_shape=kernel_shape
+            )
 
         if self._r2_inpaint_thresh is not None:
             # fill/inpaint low R2 and negative gain areas in the offset parameter
@@ -369,24 +362,25 @@ class KernelModel:
 
         return param_ra
 
-    def _full_coverage_mask(self, in_mask_ra, param_ra):
+    def _full_coverage_mask(self, in_mask_ra: RasterArray, param_ra: RasterArray) -> RasterArray:
         """
-        Helper function to create a full coverage mask i.e. a mask of output pixels fully covered by source/reference
+        Utility function to create a full coverage mask i.e. a mask of output pixels fully covered by source/reference
         data, sliding (model) kernels, and resampling kernels.
 
-        Note: that this is a strict approach that avoids any kind of partial coverage.
+        Note: this is a strict approach that avoids any kind of partial coverage.
 
         Parameters
         ----------
         in_mask_ra: RasterArray
-            An initial mask of valid source/reference pixels, as a RasterArray with dtype=='uint8'.
+            Initial mask of valid source/reference pixels, as a RasterArray with dtype=='uint8'.
         param_ra: RasterArray
-            A parameter RasterArray as returned by fit() i.e. a RasterArray in the CRS and grid corresponding to the
-            'proc_crs' attribute, whose mask corresponds to the combined reference & source masks.
+            A parameter RasterArray as returned by :meth:`fit` i.e. a RasterArray in the CRS and grid corresponding
+            to :attr:`proc_crs`, whose mask corresponds to the combined reference & source masks.
+
         Returns
         -------
-        mask_ra: RasterArray
-            The full coverage mask as a RasterArray with dtype=='uint8'
+        RasterArray
+            Full coverage mask as a RasterArray with dtype=='uint8'
         """
 
         # re-project the initial mask into proc_crs (the CRS and grid corresponding to the proc_crs attribute)
@@ -400,73 +394,67 @@ class KernelModel:
         # Similar to the block overlap amount, this removes ceil(kernel_shape/2) pixels from the nodata edge.  Note,
         # that this is the strict approach for proc_crs == ref, it could be floor(kernel_shape/2) for proc_crs == src,
         # which avoids the additional upsampling step.
-        se = cv.getStructuringElement(cv.MORPH_RECT, tuple(self._kernel_shape[::-1] + 2))
+        se = cv.getStructuringElement(cv.MORPH_RECT, tuple(np.array(self._kernel_shape[::-1]) + 2))
         mask_ra.array = cv.erode(mask, se, borderType=cv.BORDER_CONSTANT, borderValue=0)
         return mask_ra
 
-    def fit(self, ref_ra, src_ra, kernel_shape=None):
+    def fit(self, src_ra: RasterArray, ref_ra: RasterArray) -> RasterArray:
         """
         Fit sliding kernel models to reference and source blocks.
 
         Parameters
         ----------
-        ref_ra : RasterArray
+        src_ra: RasterArray
+            Source data block in a RasterArray, with the same CRS, shape & extents as ``ref_ra``.
+        ref_ra: RasterArray
             Reference data block in a RasterArray.
-        src_ra : RasterArray
-            Source data block in a RasterArray, with the same CRS, shape & extents as ref_ra.
-        kernel_shape : tuple, optional
-            Sliding kernel [height, width] in pixels.
 
         Returns
         -------
-        param_ra :RasterArray
+        RasterArray
             RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
-            R2 for each kernel model in the third band when param_image==True.
+            *R*\\ :sup:`2` for each kernel model in the third band when :attr:`find_r2` is True.
         """
-        # TODO: include a CRS comparison below i.e. one that is faster that rasterio's current implementation
-        if ((ref_ra.transform != src_ra.transform) or (ref_ra.shape != src_ra.shape)):
+        # TODO : include a CRS comparison below i.e. one that is faster that rasterio's current implementation
+        if (ref_ra.transform != src_ra.transform) or (ref_ra.shape != src_ra.shape):
             raise ValueError("'ref_ra' and 'src_ra' must have the same CRS, transform and shape")
 
-        if kernel_shape is None:
-            kernel_shape = self._kernel_shape
-        kernel_shape = tuple(kernel_shape)  # force to tuple for opencv
-
-        if self._method == Method.gain:
-            param_ra = self._fit_gain(ref_ra, src_ra, kernel_shape=kernel_shape)
-        elif self._method == Method.gain_blk_offset:
-            param_ra = self._fit_gain_blk_offset(ref_ra, src_ra, kernel_shape=kernel_shape)
+        if self._model == Model.gain:
+            param_ra = self._fit_gain(src_ra, ref_ra, kernel_shape=self._kernel_shape)
+        elif self._model == Model.gain_blk_offset:
+            param_ra = self._fit_gain_blk_offset(src_ra, ref_ra, kernel_shape=self._kernel_shape)
         else:
-            param_ra = self._fit_gain_offset(ref_ra, src_ra, kernel_shape=kernel_shape)
+            param_ra = self._fit_gain_offset(src_ra, ref_ra, kernel_shape=self._kernel_shape)
 
         return param_ra
 
-    def apply(self, src_ra, param_ra):
+    def apply(self, src_ra: RasterArray, param_ra: RasterArray) -> RasterArray:
         """
         Apply kernel models to a source block.
 
         Parameters
         ----------
-        src_ra : RasterArray
+        src_ra: RasterArray
             Source data block in a RasterArray.
-        param_ra :RasterArray
+        param_ra: RasterArray
             RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and the same
             CRS, shape & extents as src_ra.
 
         Returns
         -------
-        out_ra :RasterArray
-            Homogenised block in a RasterArray.
+        corr_ra :RasterArray
+            Corrected block in a RasterArray.
         """
-        if ((param_ra.transform != src_ra.transform) or (param_ra.shape != src_ra.shape)):
+        if (param_ra.transform != src_ra.transform) or (param_ra.shape != src_ra.shape):
             raise ValueError("'param_ra' and 'src_ra' must have the same CRS, transform and shape")
-        out_array = (param_ra.array[0] * src_ra.array) + param_ra.array[1]
-        out_ra = RasterArray.from_profile(out_array, param_ra.profile)
-        return out_ra
+        corr_array = (param_ra.array[0] * src_ra.array) + param_ra.array[1]
+        corr_ra = RasterArray.from_profile(corr_array, param_ra.profile)
+        return corr_ra
 
 
 class RefSpaceModel(KernelModel):
     """
-    A KernelModel subclass, for estimating model parmaeters in the reference image CRS and grid.
+    A KernelModel subclass, for estimating model parameters in the reference image CRS and grid.
 
     The source image block is re-projected into the reference CRS to estimate the parameters.  Estimated parameters are
     subsequently re-projected to the source CRS for application to the (original) source image block.
@@ -474,49 +462,15 @@ class RefSpaceModel(KernelModel):
     Recommended for the most common use case where the reference image has a lower resolution than the source image.
     """
 
-    def fit(self, ref_ra, src_ra, **kwargs):
-        """
-        Fit sliding kernel models to reference and source blocks.
-
-        Parameters
-        ----------
-        ref_ra : RasterArray
-            Reference data block in a RasterArray.
-        src_ra : RasterArray
-            Source data block in a RasterArray.  The src_ra bounds should that contain those of ref_ra.
-            Note that in the interests of speed, this is assumed and not checked for.
-
-        Returns
-        -------
-        param_ra :RasterArray
-            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
-            R2 for each kernel model in the third band when param_image==True.
-        """
+    def fit(self, src_ra: RasterArray, ref_ra: RasterArray) -> RasterArray:
         # choose resampling method based on whether we are up- or downsampling
         resampling = self._get_resampling(src_ra.res, ref_ra.res)
         # downsample src_ra to reference CRS and grid
         src_ds_ra = src_ra.reproject(**ref_ra.proj_profile, resampling=resampling)
         # call base class fit with reference and source RasterArrays in the reference CRS & grid
-        return KernelModel.fit(self, ref_ra, src_ds_ra, kernel_shape=self._kernel_shape)
+        return KernelModel.fit(self, src_ds_ra, ref_ra)
 
     def apply(self, src_ra, param_ra):
-        """
-        Apply kernel models to a source block.
-
-        Parameters
-        ----------
-        src_ra : RasterArray
-            Source data block in a RasterArray.
-        param_ra :RasterArray
-            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second.  The bounds
-            of param_ra should contain those of src_ra.  Note that in the interests of speed, this is assumed and not
-            checked for.
-
-        Returns
-        -------
-        out_ra :RasterArray
-            Homogenised block in a RasterArray.
-        """
         # create a parameter RasterArray containing only the first two bands of param_ra
         # (to speed up the re-projection below)
         _param_ra = RasterArray.from_profile(param_ra.array[:2], param_ra.profile)
@@ -540,7 +494,7 @@ class RefSpaceModel(KernelModel):
 
 class SrcSpaceModel(KernelModel):
     """
-    A KernelModel subclass, for estimating model parmaeters in the source image CRS and grid.
+    A KernelModel subclass, for estimating model parameters in the source image CRS and grid.
 
     The reference image block is re-projected into the source CRS to estimate the parameters.  Estimated parameters are
     subsequently applied (directly) to the source image block.
@@ -548,24 +502,7 @@ class SrcSpaceModel(KernelModel):
     Recommended for the unusual use case where the source image has a lower resolution than the reference image.
     """
 
-    def fit(self, ref_ra, src_ra, **kwargs):
-        """
-        Fit sliding kernel models to reference and source blocks.
-
-        Parameters
-        ----------
-        ref_ra : RasterArray
-            Reference data block in a RasterArray.
-        src_ra : RasterArray
-            Source data block in a RasterArray.  The ref_ra bounds should contain those of src_ra.
-            Note that in the interests of speed, this is assumed and not checked for.
-
-        Returns
-        -------
-        param_ra :RasterArray
-            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
-            R2 for each kernel model in the third band when param_image==True.
-        """
+    def fit(self, src_ra: RasterArray, ref_ra: RasterArray) -> RasterArray:
         # choose resampling method based on whether we are up- or downsampling
         resampling = self._get_resampling(ref_ra.res, src_ra.res)
         # upsample ref_ra to the source CRS and grid
@@ -573,7 +510,7 @@ class SrcSpaceModel(KernelModel):
 
         _src_ra = src_ra.copy()  # copy the source to avoid in-place changes in fit() below
         # call base class fit with source and parameter RasterArrays in the source CRS & grid
-        param_ra = KernelModel.fit(self, ref_us_ra, _src_ra, kernel_shape=self._kernel_shape)
+        param_ra = KernelModel.fit(self, _src_ra, ref_us_ra)
 
         if self._mask_partial:
             # remove r2 band from param masking
